@@ -1,24 +1,23 @@
 """Run third-party tool (e.g. :code:`mypy`) against notebook or directory."""
 import os
 import re
-import shutil
 import subprocess
 import sys
 import tempfile
 from importlib import import_module
 from pathlib import Path
 from textwrap import dedent
-from typing import Iterator, Mapping, MutableMapping, Optional, Sequence, Set, Tuple
+from typing import Iterator, MutableMapping, Optional, Sequence, Set, Tuple
 
 from pkg_resources import parse_version
 
 from nbqa import config_parser, replace_source, save_source
 from nbqa.cmdline import CLIArgs
-from nbqa.config.config import CONFIG_FILES, Configs
+from nbqa.config.config import Configs
 from nbqa.find_root import find_project_root
 from nbqa.notebook_info import NotebookInfo
 from nbqa.optional import metadata
-from nbqa.output_parser import map_python_line_to_nb_lines
+from nbqa.output_parser import Output, map_python_line_to_nb_lines
 from nbqa.text import BOLD, RESET
 
 BASE_ERROR_MESSAGE = dedent(
@@ -44,11 +43,6 @@ REPLACE_FUNCTION = {
 }
 
 
-def _hash_notebook(string: str) -> int:
-    """Hash notebook name into 8-digits."""
-    return abs(hash(string)) % (10 ** 8)
-
-
 class UnsupportedPackageVersionError(Exception):
     """Raise if installed module is older than minimum required version."""
 
@@ -136,58 +130,17 @@ def _get_all_notebooks(
     )
 
 
-def _temp_python_file_for_notebook(
-    notebook: Path, tmpdir: str, project_root: Path
-) -> Path:
-    """
-    Get temporary file to save converted notebook into.
-
-    Parameters
-    ----------
-    notebook
-        Notebook that third-party tool will be run on.
-    tmpdir
-        Temporary directory where converted notebooks will be saved.
-    project_root
-        Root of repository, where .git / .hg / .nbqa.ini file is.
-
-    Returns
-    -------
-    Path
-        Temporary Python file whose location mirrors that of the notebook, but
-        inside the temporary directory.
-
-    Raises
-    ------
-    FileNotFoundError
-        If notebook doesn't exist.
-    """
-    if not notebook.exists():
-        raise FileNotFoundError(
-            f"{BOLD}No such file or directory: {str(notebook)}{RESET}"
-        )
-    new_stem = f"{notebook.stem}_{_hash_notebook(notebook.stem)}"
-    new_parent = notebook.resolve().relative_to(project_root).parent
-    relative_notebook_path = (new_parent / new_stem).with_suffix(".py")
-    temp_python_file = Path(tmpdir) / relative_notebook_path
-    temp_python_file.parent.mkdir(parents=True, exist_ok=True)
-    return temp_python_file
-
-
 def _replace_temp_python_file_references_in_out_err(
-    tmpdirname: str,
-    temp_python_file: Path,
+    temp_python_file: str,
     notebook: Path,
     out: str,
     err: str,
-) -> Tuple[str, str]:
+) -> Output:
     """
     Replace references to temporary Python file with references to notebook.
 
     Parameters
     ----------
-    tmpdirname
-        Temporary directory used for converting notebooks to python files
     temp_python_file
         Temporary Python file where notebook was converted to.
     notebook
@@ -199,190 +152,42 @@ def _replace_temp_python_file_references_in_out_err(
 
     Returns
     -------
-    out
-        Stdout with temporary directory replaced by current working directory.
-    err
-        Stderr with temporary directory replaced by current working directory.
+    Output
+        Stdout, stderr with temporary directory replaced by current working directory.
     """
-    # 1. Relative path is used because some tools like pylint always report only
-    # the relative path of the file(relative to project root),
-    # though absolute path was passed as the input.
-    # 2. This `resolve()` part is necessary to handle cases when the path used
-    # is a symlink as well as no normalize the path.
-    # I couldn't reproduce this locally, but during CI, on the Windows job, I found
-    # that VSSADM~1 was changing into VssAdministrator.
-    paths = (
-        str(path)
-        for path in [
-            temp_python_file,
-            temp_python_file.resolve(),
-            temp_python_file.relative_to(tmpdirname),
-        ]
-    )
+    basename = os.path.basename(temp_python_file)
+    out = out.replace(basename, str(notebook.stem) + ".ipynb")
+    err = err.replace(basename, str(notebook.stem) + ".ipynb")
 
-    notebook_path = str(notebook)
-    for path in paths:
-        out = out.replace(path, notebook_path)
-        err = err.replace(path, notebook_path)
+    suffix = ".py"
+    out = out.replace(basename[: -len(suffix)], notebook.stem)
+    err = err.replace(basename[: -len(suffix)], notebook.stem)
 
-    out = out.replace(f"{tmpdirname}{os.sep}", "")
-    err = err.replace(f"{tmpdirname}{os.sep}", "")
-
-    out = out.replace(temp_python_file.stem, notebook.stem)
-    err = err.replace(temp_python_file.stem, notebook.stem)
-
-    return out, err
+    return Output(out, err)
 
 
-def _create_blank_init_files(
-    notebook: Path, tmpdirname: str, project_root: Path
-) -> None:
-    """
-    Replicate local (possibly blank) __init__ files to temporary directory.
-
-    Parameters
-    ----------
-    notebook
-        Notebook third-party tool is being run against.
-    tmpdirname
-        Temporary directory to store converted notebooks in.
-    project_root
-        Root of repository, where .git / .hg / .nbqa.ini file is.
-    """
-    parts = notebook.resolve().relative_to(project_root).parts
-
-    for idx in range(1, len(parts)):
-        init_file = Path(os.path.join(*parts[:idx])) / "__init__.py"
-        Path(tmpdirname).joinpath(init_file).parent.mkdir(parents=True, exist_ok=True)
-        Path(tmpdirname).joinpath(init_file).touch()
-
-
-def _preserve_config_files(
-    config_files: Sequence[str], tmpdirname: str, project_root: Path
-) -> None:
-    """
-    Copy local config file to temporary directory.
-
-    Parameters
-    ----------
-    config_files
-        Config files for third-party tool (e.g. mypy).
-    tmpdirname
-        Temporary directory to store converted notebooks in.
-    project_root
-        Root of repository, where .git / .hg / .nbqa.ini file is.
-    """
-    for config_file in config_files:
-        config_file_path = project_root / config_file
-        if config_file_path.exists():
-            target_location = Path(tmpdirname) / config_file_path.resolve().relative_to(
-                project_root
-            )
-            target_location.parent.mkdir(parents=True, exist_ok=True)
-            shutil.copy(
-                str(config_file_path),
-                str(target_location),
-            )
-
-
-def _get_arg(
-    root_dir: str,
-    tmpdirname: str,
-    nb_to_py_mapping: Mapping[Path, Path],
-    project_root: Path,
-) -> Path:
-    """
-    Get argument to run command against.
-
-    If running against a single notebook, it'll be the filepath of the converted
-    notebook in the temporary directory.
-    If running against a directory, it'll be the directory mirrored in the temporary
-    directory.
-
-    Parameters
-    ----------
-    root_dir
-        Notebook or directory third-party tool is being run against.
-    tmpdirname
-        Temporary directory where converted notebooks are stored.
-    nb_to_py_mapping
-        Mapping between notebooks and Python files corresponding to converted notebooks.
-    project_root
-        Root of repository, where .git / .hg / .nbqa.ini file is.
-
-    Returns
-    -------
-    Path
-        Notebook or directory to run third-party tool against.
-
-    Examples
-    --------
-    >>> root_dir = "root_dir"
-    >>> tmpdirname = "tmpdir"
-    >>> nb_to_py_mapping = {
-    ...     Path('my_notebook.ipynb'): Path('tmpdir').joinpath('my_notebook.py')
-    ... }
-    >>> _get_arg(root_dir, tmpdirname, nb_to_py_mapping).as_posix()
-    'tmpdir/my_notebook.py'
-    """
-    if Path(root_dir).is_dir():
-        return Path(tmpdirname) / Path(root_dir).resolve().relative_to(project_root)
-    return nb_to_py_mapping[Path(root_dir)]
-
-
-def _get_all_args(
-    root_dirs: Sequence[str],
-    tmpdirname: str,
-    nb_to_py_mapping: Mapping[Path, Path],
-    project_root: Path,
-) -> Sequence[Path]:
-    """
-    Get all arguments to run command against.
-
-    Parameters
-    ----------
-    root_dirs
-        All notebooks or directories third-party tool is being run against.
-    tmpdirname
-        Temporary directory where converted notebooks are stored.
-    nb_to_py_mapping
-        Mapping between notebooks and Python files corresponding to converted notebooks.
-    project_root
-        Root of repository, where .git / .hg / .nbqa.ini file is.
-
-    Returns
-    -------
-    Sequence[Path]
-        All notebooks or directories to run third-party tool against.
-    """
-    return [_get_arg(i, tmpdirname, nb_to_py_mapping, project_root) for i in root_dirs]
-
-
-def _get_mtimes(arg: Path) -> Set[float]:
+def _get_mtimes(arg: str) -> Set[float]:
     """
     Get the modification times of any converted notebooks.
 
     Parameters
     ----------
     arg
-        Notebook or directory to run 3rd party tool on.
+        Notebook to run 3rd party tool on.
 
     Returns
     -------
     Set
         Modification times of any converted notebooks.
     """
-    if not arg.is_dir():
-        return {os.path.getmtime(str(arg))}
-    return {os.path.getmtime(str(i)) for i in arg.rglob("*.py")}
+    return {os.path.getmtime(arg)}
 
 
 def _run_command(
     command: str,
-    tmpdirname: str,
     cmd_args: Sequence[str],
-    args: Sequence[Path],
-) -> Tuple[str, str, int, bool]:
+    args: Sequence[str],
+) -> Tuple[Output, int, bool]:
     """
     Run third-party tool against given file or directory.
 
@@ -390,8 +195,6 @@ def _run_command(
     ----------
     command
         Third-party tool (e.g. :code:`mypy`) to run against notebook.
-    tmpdirname
-        Temporary directory where converted notebooks will be stored.
     cmd_args
         Flags to pass to third-party tool (e.g. :code:`--verbose`).
     args
@@ -399,10 +202,8 @@ def _run_command(
 
     Returns
     -------
-    out
-        Captured stdout from running third-party tool.
-    err
-        Captured stderr from running third-party tool.
+    output
+        Captured stdout, stderr from running third-party tool.
     output_code
         Return code from third-party tool.
     mutated
@@ -413,23 +214,12 @@ def _run_command(
     ValueError
         If third-party tool isn't found in system.
     """
-    env = os.environ.copy()
-    env[
-        "MYPYPATH"
-    ] = f"{env.get('MYPYPATH', '').rstrip(os.pathsep)}{os.pathsep}{os.getcwd()}"
-
-    env[
-        "PYTHONPATH"
-    ] = f"{env.get('PYTHONPATH', '').rstrip(os.pathsep)}{os.pathsep}{os.getcwd()}"
-
     before = [_get_mtimes(i) for i in args]
 
     output = subprocess.run(
-        [sys.executable, "-m", command, *(str(i) for i in args), *cmd_args],
+        [sys.executable, "-m", command, *args, *cmd_args],
         stderr=subprocess.PIPE,
         stdout=subprocess.PIPE,
-        cwd=tmpdirname,
-        env=env,
         universal_newlines=True,  # from Python3.7 this can be replaced with `text`
     )
 
@@ -440,7 +230,7 @@ def _run_command(
     out = output.stdout
     err = output.stderr
 
-    return out, err, output_code, mutated
+    return Output(out, err), output_code, mutated
 
 
 def _get_command_not_found_msg(command: str) -> str:
@@ -501,8 +291,8 @@ def _get_configs(cli_args: CLIArgs, project_root: Path) -> Configs:
     return cli_config.merge(Configs.get_default_config(cli_args.command))
 
 
-def _run_on_one_root_dir(
-    cli_args: CLIArgs, configs: Configs, project_root: Path
+def _main(  # pylint: disable=R0912,R0914,R0911
+    cli_args: CLIArgs, configs: Configs
 ) -> int:
     """
     Run third-party tool on a single notebook or directory.
@@ -513,30 +303,27 @@ def _run_on_one_root_dir(
         Commanline arguments passed to nbqa.
     configs
         Configuration passed to nbqa from commandline or via a config file
-    project_root
-        Root of repository, where .git / .hg / .nbqa.ini file is.
 
     Returns
     -------
     int
         Output code from third-party tool.
-
-    Raises
-    ------
-    RuntimeError
-        If unable to parse or reconstruct notebook.
-    SystemExit
-        If third-party tool would've reformatted notebook but ``--nbqa-mutate``
-        wasn't passed.
     """
-    with tempfile.TemporaryDirectory() as tmpdirname:
+    nb_to_py_mapping = {}
+    for notebook in _get_all_notebooks(
+        cli_args.root_dirs, configs.nbqa_files, configs.nbqa_exclude
+    ):
+        if not notebook.exists():
+            sys.stderr.write(f"{BOLD}No such file or directory: {str(notebook)}{RESET}")
+            return 1
 
-        nb_to_py_mapping = {
-            notebook: _temp_python_file_for_notebook(notebook, tmpdirname, project_root)
-            for notebook in _get_all_notebooks(
-                cli_args.root_dirs, configs.nbqa_files, configs.nbqa_exclude
-            )
-        }
+        nb_to_py_mapping[notebook] = tempfile.mkstemp(
+            dir=str(notebook.parent),
+            prefix=notebook.stem,
+            suffix=".py",
+        )
+
+    try:  # pylint disable=R0912
 
         if not nb_to_py_mapping:
             sys.stderr.write(
@@ -545,57 +332,48 @@ def _run_on_one_root_dir(
             )
             return 0
 
-        config_files = (
-            [configs.nbqa_config]
-            if configs.nbqa_config
-            else CONFIG_FILES[cli_args.command]
-        )
-        _preserve_config_files(config_files, tmpdirname, project_root)
-
         nb_info_mapping: MutableMapping[Path, NotebookInfo] = {}
 
-        for notebook, temp_python_file in nb_to_py_mapping.items():
+        for notebook, (file_descriptor, _) in nb_to_py_mapping.items():
             try:
                 nb_info_mapping[notebook] = save_source.main(
                     notebook,
-                    temp_python_file,
+                    file_descriptor,
                     configs.nbqa_process_cells,
                     cli_args.command,
                 )
-            except Exception as exc:
-                raise RuntimeError(
+            except Exception:  # pylint: disable=W0703
+                sys.stderr.write(
                     BASE_ERROR_MESSAGE.format(f"Error parsing {str(notebook)}")
-                ) from exc
+                )
+                return 1
 
-            _create_blank_init_files(notebook, tmpdirname, project_root)
-
-        out, err, output_code, mutated = _run_command(
+        output, output_code, mutated = _run_command(
             cli_args.command,
-            tmpdirname,
             configs.nbqa_addopts,
-            _get_all_args(
-                cli_args.root_dirs, tmpdirname, nb_to_py_mapping, project_root
-            ),
+            [i[1] for i in nb_to_py_mapping.values()],
         )
 
-        for notebook, temp_python_file in nb_to_py_mapping.items():
-            out, err = _replace_temp_python_file_references_in_out_err(
-                tmpdirname, temp_python_file, notebook, out, err
+        actually_mutated = False
+        for notebook, (_, temp_python_file) in nb_to_py_mapping.items():
+            output = _replace_temp_python_file_references_in_out_err(
+                temp_python_file, notebook, output.out, output.err
             )
             try:
-                out, err = map_python_line_to_nb_lines(
+                output = map_python_line_to_nb_lines(
                     cli_args.command,
-                    out,
-                    err,
+                    output.out,
+                    output.err,
                     notebook,
                     nb_info_mapping[notebook].cell_mappings,
                 )
-            except Exception as exc:  # pylint: disable=W0703
+            except Exception as exc:  # pylint: disable=W0703  # pragma: nocover
                 msg = (
                     f"{repr(exc)} while parsing output "
                     f"from applying {cli_args.command} to {str(notebook)}"
                 )
                 sys.stderr.write(BASE_ERROR_MESSAGE.format(msg))
+                return 1
 
             if mutated:
                 if not configs.nbqa_mutate and not configs.nbqa_diff:
@@ -612,23 +390,32 @@ def _run_on_one_root_dir(
                         """
                     )
                     # pylint: enable=C0301
-                    raise SystemExit(msg)
+                    sys.stderr.write(msg)
+                    return 1
 
                 try:
-                    REPLACE_FUNCTION[configs.nbqa_diff](
-                        temp_python_file,
-                        notebook,
-                        nb_info_mapping[notebook],
+                    actually_mutated = (
+                        REPLACE_FUNCTION[configs.nbqa_diff](
+                            temp_python_file,
+                            notebook,
+                            nb_info_mapping[notebook],
+                        )
+                        or actually_mutated
                     )
-                except Exception as exc:
-                    raise RuntimeError(
+                except Exception:  # pylint: disable=W0703
+                    sys.stderr.write(
                         BASE_ERROR_MESSAGE.format(
                             f"Error reconstructing {str(notebook)}"
                         )
-                    ) from exc
+                    )
+                    return 1
 
-        sys.stdout.write(out)
-        sys.stderr.write(err)
+        sys.stdout.write(output.out)
+        sys.stderr.write(output.err)
+
+        if mutated and not actually_mutated:
+            output_code = 0
+            mutated = False
 
         if configs.nbqa_diff:
             if mutated:
@@ -637,6 +424,15 @@ def _run_on_one_root_dir(
                 )
             return output_code
 
+    finally:
+        for file_descriptor, tmp_path in nb_to_py_mapping.values():  # pragma: nocover
+            try:
+                os.close(file_descriptor)
+            except OSError:
+                # was already closed
+                pass
+            os.remove(tmp_path)
+
     return output_code
 
 
@@ -672,7 +468,7 @@ def _check_command_is_installed(command: str) -> None:
                 )
 
 
-def main(argv: Optional[Sequence[str]] = None) -> None:
+def main(argv: Optional[Sequence[str]] = None) -> int:
     """
     Run third-party tool (e.g. :code:`mypy`) against notebook or directory.
 
@@ -688,10 +484,8 @@ def main(argv: Optional[Sequence[str]] = None) -> None:
     configs: Configs = _get_configs(cli_args, project_root)
     configs.validate()
 
-    output_code = _run_on_one_root_dir(cli_args, configs, project_root)
-
-    sys.exit(output_code)
+    return _main(cli_args, configs)
 
 
 if __name__ == "__main__":
-    main()
+    sys.exit(main())
