@@ -1,60 +1,61 @@
 """Module responsible for storing and handling nbqa configuration."""
 
-from collections import defaultdict
+import collections
 from pathlib import Path
 from shlex import split
 from textwrap import dedent
-from typing import (
-    Any,
-    Callable,
-    ClassVar,
-    DefaultDict,
-    Dict,
-    Mapping,
-    NamedTuple,
-    Optional,
-    Sequence,
-    Union,
-)
+from typing import Any, Callable, ClassVar, Dict, Mapping, Optional, Sequence, Union
 
 import toml
 from pkg_resources import resource_filename
 
 from nbqa.cmdline import CLIArgs
 
-CONFIG_FILES: DefaultDict[str, Sequence[str]] = defaultdict(
-    lambda: ["setup.cfg", "tox.ini", "pyproject.toml"]
-)
-CONFIG_FILES["autoflake"] = []
-CONFIG_FILES["black"] = ["pyproject.toml"]
-CONFIG_FILES["check-ast"] = []
-CONFIG_FILES["doctest"] = []
-CONFIG_FILES["flake8"] = ["setup.cfg", "tox.ini", ".flake8"]
-CONFIG_FILES["isort"] = [
-    ".isort.cfg",
-    "pyproject.toml",
-    "setup.cfg",
-    "tox.ini",
-    ".editorconfig",
-]
-CONFIG_FILES["mypy"] = ["mypy.ini", ".mypy.ini", "setup.cfg"]
-CONFIG_FILES["pylint"] = ["pylintrc", ".pylintrc", "pyproject.toml", "setup.cfg"]
-CONFIG_FILES["pyupgrade"] = []
-
 ConfigParser = Callable[[str], Union[str, bool, Sequence[str]]]
 
 
-class _ConfigSections(NamedTuple):  # pylint: disable=R0903
-    """Stores all the config section names."""
+class _ConfigSections(
+    collections.namedtuple(
+        "_ConfigSections",
+        (
+            "ADDOPTS",
+            "CONFIG",
+            "IGNORE_CELLS",
+            "PROCESS_CELLS",
+            "MUTATE",
+            "DIFF",
+            "FILES",
+            "EXCLUDE",
+        ),
+    )
+):
+    """Config sections."""
 
-    ADDOPTS: str = "addopts"
-    CONFIG: str = "config"
-    IGNORE_CELLS: str = "ignore_cells"
-    PROCESS_CELLS: str = "process_cells"
-    MUTATE: str = "mutate"
-    DIFF: str = "diff"
-    FILES: str = "files"
-    EXCLUDE: str = "exclude"
+    __slots__ = ()
+
+    def __new__(  # pylint: disable=R0913
+        cls,
+        ADDOPTS: str = "addopts",
+        CONFIG: str = "config",
+        IGNORE_CELLS: str = "ignore_cells",
+        PROCESS_CELLS: str = "process_cells",
+        MUTATE: str = "mutate",
+        DIFF: str = "diff",
+        FILES: str = "files",
+        EXCLUDE: str = "exclude",
+    ) -> "_ConfigSections":
+        """Python3.6.0 doesn't support defaults for namedtuples."""
+        return super().__new__(
+            cls,
+            ADDOPTS,
+            CONFIG,
+            IGNORE_CELLS,
+            PROCESS_CELLS,
+            MUTATE,
+            DIFF,
+            FILES,
+            EXCLUDE,
+        )
 
 
 CONFIG_SECTIONS = _ConfigSections()
@@ -87,21 +88,21 @@ class Configs:
         Global file exclude pattern.
     """
 
-    _config_section_parsers: ClassVar[Dict[str, ConfigParser]] = {}
-    _config_section_parsers[CONFIG_SECTIONS.ADDOPTS] = (
+    CONFIG_SECTION_PARSERS: ClassVar[Dict[str, ConfigParser]] = {}
+    CONFIG_SECTION_PARSERS[CONFIG_SECTIONS.ADDOPTS] = (
         lambda arg: split(arg) if isinstance(arg, str) else arg
     )
-    _config_section_parsers[CONFIG_SECTIONS.CONFIG] = str
-    _config_section_parsers[CONFIG_SECTIONS.IGNORE_CELLS] = (
+    CONFIG_SECTION_PARSERS[CONFIG_SECTIONS.CONFIG] = str
+    CONFIG_SECTION_PARSERS[CONFIG_SECTIONS.IGNORE_CELLS] = (
         lambda arg: arg.split(",") if isinstance(arg, str) else arg
     )
-    _config_section_parsers[CONFIG_SECTIONS.PROCESS_CELLS] = (
+    CONFIG_SECTION_PARSERS[CONFIG_SECTIONS.PROCESS_CELLS] = (
         lambda arg: arg.split(",") if isinstance(arg, str) else arg
     )
-    _config_section_parsers[CONFIG_SECTIONS.MUTATE] = bool
-    _config_section_parsers[CONFIG_SECTIONS.DIFF] = bool
-    _config_section_parsers[CONFIG_SECTIONS.FILES] = str
-    _config_section_parsers[CONFIG_SECTIONS.EXCLUDE] = str
+    CONFIG_SECTION_PARSERS[CONFIG_SECTIONS.MUTATE] = bool
+    CONFIG_SECTION_PARSERS[CONFIG_SECTIONS.DIFF] = bool
+    CONFIG_SECTION_PARSERS[CONFIG_SECTIONS.FILES] = str
+    CONFIG_SECTION_PARSERS[CONFIG_SECTIONS.EXCLUDE] = str
 
     _mutate: bool = False
     _config: Optional[str] = None
@@ -124,7 +125,7 @@ class Configs:
             Config value
         """
         if value:
-            self.__setattr__(f"_{config}", self._config_section_parsers[config](value))
+            self.__setattr__(f"_{config}", self.CONFIG_SECTION_PARSERS[config](value))
 
     @property
     def nbqa_mutate(self) -> bool:
