@@ -4,7 +4,9 @@ Replace :code:`source` code cells of original notebook with ones from converted
 The converted file will have had the third-party tool run against it by now.
 """
 
+import copy
 import json
+import os
 import sys
 from difflib import unified_diff
 from shutil import move
@@ -19,6 +21,8 @@ from typing import (
     Set,
 )
 
+import tokenize_rt
+
 from nbqa.handle_magics import MagicHandler
 from nbqa.notebook_info import NotebookInfo
 from nbqa.save_source import CODE_SEPARATOR
@@ -52,11 +56,24 @@ def _restore_semicolon(
     -------
     str
         New source with removed semicolon restored.
-    """
-    rstripped_source = source.rstrip()
-    if cell_number in trailing_semicolons and not rstripped_source.endswith(";"):
-        source = rstripped_source + ";"
 
+    Raises
+    ------
+    AssertionError
+        If code thought to be unreachable is reached.
+    """
+    if cell_number in trailing_semicolons:
+        tokens = tokenize_rt.src_to_tokens(source)
+        for idx, token in tokenize_rt.reversed_enumerate(tokens):
+            if not token.src.strip(" \n") or token.name == "COMMENT":
+                continue
+            tokens[idx] = token._replace(src=token.src + ";")
+            break
+        else:  # pragma: nocover
+            raise AssertionError(
+                "Unreachable code, please report bug at https://github.com/nbQA-dev/nbQA/issues"
+            )
+        source = tokenize_rt.tokens_to_src(tokens)
     return source
 
 
@@ -115,7 +132,7 @@ def _get_new_source(
     )
 
 
-def _get_pycells(python_file: "Path") -> Iterator[str]:
+def _get_pycells(python_file: str) -> Iterator[str]:
     """
     Parse cells from Python file.
 
@@ -129,7 +146,8 @@ def _get_pycells(python_file: "Path") -> Iterator[str]:
     Iterator
         Parsed cells.
     """
-    txt = python_file.read_text()
+    with open(python_file, encoding="utf-8") as handle:
+        txt = handle.read()
     if txt.startswith(CODE_SEPARATOR):
         txt = txt[len(CODE_SEPARATOR) :]
     return iter(txt.split(CODE_SEPARATOR))
@@ -156,7 +174,7 @@ def _notebook_code_cells(
             yield cell
 
 
-def mutate(python_file: "Path", notebook: "Path", notebook_info: NotebookInfo) -> None:
+def mutate(python_file: str, notebook: "Path", notebook_info: NotebookInfo) -> bool:
     """
     Replace :code:`source` code cells of original notebook.
 
@@ -168,8 +186,14 @@ def mutate(python_file: "Path", notebook: "Path", notebook_info: NotebookInfo) -
         Jupyter Notebook third-party tool is run against (unmodified).
     notebook_info
         Information about notebook cells used for processing
+
+    Returns
+    -------
+    bool
+        Whether mutation actually happened.
     """
     notebook_json = json.loads(notebook.read_text(encoding="utf-8"))
+    original_notebook_json = copy.deepcopy(notebook_json)
 
     pycells = _get_pycells(python_file)
     for code_cell_number, cell in enumerate(
@@ -179,14 +203,17 @@ def mutate(python_file: "Path", notebook: "Path", notebook_info: NotebookInfo) -
             continue
         cell["source"] = _get_new_source(code_cell_number, notebook_info, next(pycells))
 
-    temp_notebook = python_file.parent / notebook.name
-    temp_notebook.write_text(
-        f"{json.dumps(notebook_json, indent=1, ensure_ascii=False)}\n", encoding="utf-8"
-    )
+    if original_notebook_json == notebook_json:
+        return False
+
+    temp_notebook = os.path.join(os.path.dirname(python_file), notebook.name)
+    with open(temp_notebook, "w", encoding="utf-8") as handle:
+        handle.write(f"{json.dumps(notebook_json, indent=1, ensure_ascii=False)}\n")
     move(str(temp_notebook), str(notebook))
+    return True
 
 
-def _print_diff(code_cell_number: int, cell_diff: Iterator[str]) -> None:
+def _print_diff(code_cell_number: int, cell_diff: Iterator[str]) -> bool:
     """
     Print diff between cells, colouring as appropriate.
 
@@ -196,6 +223,11 @@ def _print_diff(code_cell_number: int, cell_diff: Iterator[str]) -> None:
         Current cell number
     cell_diff
         Diff between original and new versions of cell.
+
+    Returns
+    -------
+    bool
+        Whether non-null diff was printed.
     """
     line_changes = []
     for line in cell_diff:
@@ -212,9 +244,11 @@ def _print_diff(code_cell_number: int, cell_diff: Iterator[str]) -> None:
         header = f"Cell {code_cell_number}"
         headers = [f"{BOLD}{header}{RESET}\n", f"{'-'*len(header)}\n"]
         sys.stdout.writelines(headers + line_changes + ["\n"])
+        return True
+    return False
 
 
-def diff(python_file: "Path", notebook: "Path", notebook_info: NotebookInfo) -> None:
+def diff(python_file: str, notebook: "Path", notebook_info: NotebookInfo) -> bool:
     """
     View diff between new source of code cells and original sources.
 
@@ -226,11 +260,18 @@ def diff(python_file: "Path", notebook: "Path", notebook_info: NotebookInfo) ->
         Jupyter Notebook third-party tool is run against (unmodified).
     notebook_info
         Information about notebook cells used for processing
+
+    Returns
+    -------
+    bool
+        Whether non-null diff was produced.
     """
     notebook_json = json.loads(notebook.read_text(encoding="utf-8"))
 
     pycells = _get_pycells(python_file)
 
+    actually_mutated = False
+
     for code_cell_number, cell in enumerate(
         _notebook_code_cells(notebook_json), start=1
     ):
@@ -245,4 +286,5 @@ def diff(python_file: "Path", notebook: "Path", notebook_info: NotebookInfo) ->
             fromfile=str(notebook),
             tofile=str(notebook),
         )
-        _print_diff(code_cell_number, cell_diff)
+        actually_mutated = _print_diff(code_cell_number, cell_diff) or actually_mutated
+    return actually_mutated
